#include <stdio.h>
#include <glib.h>
#include <string.h>
#include <stdlib.h>
#include "graph.h"
#include "causalito.h"
#include "model.h"
#include "causalgraph.h"
#include "maxflow.h"
#include "misc.h"
#include "globals.h"

static Graph bg, GC0, GprimeC0;
GSList * E0, * V0, * C, * Cstar, * C0, * EC, * VC, * ECstar, * VCstar, * VN, * EbarreC, * VbarreC, * E, * V, * VprimeC, * EbarreCconds, * tmpVlist, * GammaVprimeC;

/**
 * compute_causal_graph: Compute the causal graph of the model
 * @void: 
 * 
 * 
 * 
 * Return value: 
 **/
int compute_causal_graph(void)
{
  int mode, i, j, ngc0, ntmpcg;
  char tmpstr[1024], * ptrstr;
  Graph tmpcg;
  GSList * conditions_before, * conditions_now, * GammaVprimeCUVC, * EGamma;
  Equation e;
  Arc a, a1, unarc;
  
  VERBMSG("Computing causal graph... Starting.\n");

  /********************
   * Algorithm STEP 1 *
   ********************/
  VERBMSG("Causalito algorithm: step 1.\n")
  E0 = NULL;
  V0 = NULL;
  if (model->nconds == 0) { /* no conditions, only one operating mode */
    E0 = model->eqs_name_list;
    V0 = model->vars_name_list;
    VERBMSG("First and uniq operating mode.\n");
  } else {
    conditions_now = NULL;
    conditions_now = get_conditions_in_nth_mode(conditions_now, 1);
    VERBMSG("1st operating mode: %s.\n",
	    sprint_char_list(tmpstr, conditions_now));
    conditions_now = g_slist_append(conditions_now, "t");
    E0 = get_equations_with_conditions(E0, conditions_now);
    V0 = get_variables_in_equations(V0, E0);
  }

  bg = (Graph) malloc(sizeof(struct _Graph));
  bg->nodes_list = NULL;
  bg->arcs_list = NULL;

  i = g_slist_length(E0);
  j = g_slist_length(V0);
  if (i != j) {
    BUG("It's a bad idea to compute a perfert matching between %d equations and %d variables. Causalito cannot deal with not self-contained systems.\n", i, j);
  }
      
  make_bipgraph(E0, V0, bg);
  max_flow(bg, SOURCE, SINK);
  check_flow_in_graph(bg);
/*   DEBUG("Liste E0: [%d] %s\n", i, sprint_char_list(tmpstr, E0)); */
/*   DEBUG("Liste V0: [%d] %s\n", j, sprint_char_list(tmpstr, V0)); */
/*   debug_print_graph(bg); */
  remember_matched_var(E0, V0, bg);
/*   for (i = 0 ; i < model->neqs ; i++) { */
/*     e = g_slist_nth_data(model->eqs_list, i); */
/*     if (e->matched) { */
/*       DEBUG("%s matches %s\n", e->eq, e->matched); */
/*     } else { */
/*             DEBUG("%s has no match.\n", e->eq); */
/*     } */
/*   } */
    
      
  GC0 = (Graph) malloc(sizeof(struct _Graph));
  GC0->nodes_list = NULL;
  GC0->arcs_list = NULL;

  bipg_to_causalg(bg, GC0, E0);

/*   debug_print_graph(tmpcg); */
  
  for (mode = 2 ; mode <= model->nconfs ; mode++) {

    /********************
     * Algorithm STEP 2 *
     ********************/
    VERBMSG("Causalito algorithm: step 2.\n")

    conditions_before = NULL;
    conditions_now = NULL;
    conditions_before = get_conditions_in_nth_mode(conditions_before,
						   mode - 1);
    conditions_now = get_conditions_in_nth_mode(conditions_now, mode);
    C = NULL;
    C = remove_elts_from(conditions_before, conditions_now, C);
    Cstar = NULL;
    Cstar = remove_elts_from(conditions_now, conditions_before, Cstar);
    VERBMSG("%dth operating mode: %s.\n", mode,
	    sprint_char_list(tmpstr, conditions_now));
/*     DEBUG("conditions_before = %s\n", */
/* 	  sprint_char_list(tmpstr, conditions_before)); */
/*     DEBUG("conditions_now = %s\n", */
/* 	  sprint_char_list(tmpstr, conditions_now)); */
/*     DEBUG("C = %s\n", sprint_char_list(tmpstr, C)); */
/*     DEBUG("Cstar = %s\n", sprint_char_list(tmpstr, C)); */

    EC = NULL;
    EC = get_equations_with_conditions(EC, C);
    VC = NULL;
    j = g_slist_length(EC);
    for (i = 0 ; i < j ; i++) {
      ptrstr = g_slist_nth_data(EC, i);
      e = is_eq_in_model(ptrstr);
      if (e == NULL) {
	BUG("Cannot find equation \"%s\" in model.\n", ptrstr);
      } else {
	add_string_in_list(e->matched, VC);
      }
    }
/*     DEBUG("C = %s\n", sprint_char_list(tmpstr, C)); */
/*     DEBUG("EC = %s\n", sprint_char_list(tmpstr, EC)); */
/*     DEBUG("VC = %s\n", sprint_char_list(tmpstr, VC)); */

    ECstar = NULL;
    ECstar = get_equations_with_conditions(ECstar, Cstar);
    VCstar = NULL;
    VCstar = get_variables_in_equations(VCstar, ECstar);
/*     DEBUG("Cstar = %s\n", sprint_char_list(tmpstr, Cstar)); */
/*     DEBUG("ECstar = %s\n", sprint_char_list(tmpstr, ECstar)); */
/*     DEBUG("VCstar = %s\n", sprint_char_list(tmpstr, VCstar)); */

    /* 'EbarreC' is the set of equations neither submitted to 'C' or
       'Cstar': so we take 'conditions_now', we hade "t", we remove
       'Cstar' and 'EbarreC' is the list of equations submitted to a
       conditions appearing in the new value of 'conditions_now'. */
    EbarreC = NULL;
    EbarreCconds = NULL;
    EbarreCconds = remove_elts_from(conditions_now, Cstar, EbarreCconds);
    EbarreCconds = g_slist_append(EbarreCconds, "t");
/*     DEBUG("EbarreCconds = %s\n", sprint_char_list(tmpstr, EbarreCconds)); */
    EbarreC = get_equations_with_conditions(EbarreC, EbarreCconds);
/*     DEBUG("EbarreC = %s\n", sprint_char_list(tmpstr, EbarreC)); */
    VbarreC = NULL;
    VbarreC = get_variables_in_equations(VbarreC, EbarreC);
/*     DEBUG("VbarreC = %s\n", sprint_char_list(tmpstr, VbarreC)); */

    E = NULL;
    E = list_concat_uniq(E, EbarreC, ECstar);
/*     DEBUG("EbarreC = %s\n", sprint_char_list(tmpstr, EbarreC)); */
/*     DEBUG("ECstar = %s\n", sprint_char_list(tmpstr, ECstar)); */
/*     DEBUG("E = %s\n", sprint_char_list(tmpstr, E)); */
    if (g_slist_length(E) !=
	(g_slist_length(EbarreC) + g_slist_length(ECstar))) {
      BUG("It seems that there are common elements in EbarreC and ECstar.\n");
    }
    V = NULL;
    V = get_variables_in_equations(V, E);
    /*     DEBUG("V = %s\n", sprint_char_list(tmpstr, V)); */
    add_string_in_list("t", conditions_before);
/*     DEBUG("conditions_before = %s\n",  */
/*  	  sprint_char_list(tmpstr, conditions_before)); */
    E0 = NULL;
    E0 = get_equations_with_conditions(E0, conditions_before);
    V0 = NULL;
    V0 = get_variables_in_equations(V0, E0);
/*     DEBUG("E0 = %s\n", sprint_char_list(tmpstr, E0)); */
/*     DEBUG("V0 = %s\n", sprint_char_list(tmpstr, V0)); */

    VprimeC = NULL;
    VprimeC = list_inter_uniq(VprimeC, V, VC);
/*     DEBUG("V = %s\n", sprint_char_list(tmpstr, V)); */
/*     DEBUG("VC = %s\n", sprint_char_list(tmpstr, VC)); */
/*     DEBUG("VprimeC = %s\n", sprint_char_list(tmpstr, VprimeC)); */

    VN = NULL;
    VN = remove_elts_from(VCstar, V0, VN);
/*     DEBUG("VCstar = %s\n", sprint_char_list(tmpstr, VCstar)); */
/*     DEBUG("V0 = %s\n", sprint_char_list(tmpstr, V0)); */
/*     DEBUG("VN = %s\n", sprint_char_list(tmpstr, VN)); */

    
    DEBUG("Cstar = %s\n", sprint_char_list(tmpstr, Cstar));
    DEBUG("EC = %s\n", sprint_char_list(tmpstr, EC));
    DEBUG("VC = %s\n", sprint_char_list(tmpstr, VC));
    DEBUG("ECstar = %s\n", sprint_char_list(tmpstr, ECstar));
    DEBUG("VCstar = %s\n", sprint_char_list(tmpstr, VCstar));
    DEBUG("EbarreC = %s\n", sprint_char_list(tmpstr, EbarreC));
    DEBUG("VbarreC = %s\n", sprint_char_list(tmpstr, VbarreC));
    DEBUG("E = %s\n", sprint_char_list(tmpstr, E));
    DEBUG("V = %s\n", sprint_char_list(tmpstr, V));
    DEBUG("VN = %s\n", sprint_char_list(tmpstr, VN));
    DEBUG("VprimeC = %s\n", sprint_char_list(tmpstr, VprimeC));
    

    /**********************
     * Algorithm STEP 2.1 *
     **********************/
    VERBMSG("Causalito algorithm: step 2.1.\n");

/*     if (((g_slist_length(VprimeC) == 0) && (g_slist_length(VbarreC) == 0)) */
/* 	|| (g_slist_length(VprimeC) == 0)) { */

    if (SUCCESS) {

      /*********************************
       * Algorithm STEP 2.1, THEN part *
       *********************************/
      VERBMSG("Causalito algorithm: step 2.1, THEN part.\n");
      /* Label the arcs entering the variables of VprimeC */
      ngc0 = g_slist_length(GC0->arcs_list);
      for (i = 0 ; i < ngc0 ; i++) {
	a = g_slist_nth_data(GC0->arcs_list, i);
	if (is_string_in_list((a->end)->name, VC)) {
	  if (strlen(a->label) == 0)
	    sprintf(a->label, "(");
	  else sprintf(a->label, "%s (", a->label);
	  sprint_char_list(tmpstr, C);
	  sprintf(a->label, "%s%s)", a->label, tmpstr);
	}
      }
      tmpVlist = NULL;
      tmpVlist = list_concat_uniq(tmpVlist, VprimeC, VN);

      bg = (Graph) malloc(sizeof(struct _Graph));
      bg->nodes_list = NULL;
      bg->arcs_list = NULL;

      i = g_slist_length(ECstar); 
      j = g_slist_length(tmpVlist);
      if (i != j) {
	BUG("It's a bad idea to compute a perfert matching between %d equations and %d variables. Causalito cannot deal with not self-contained systems.\n", i, j);
      }
      make_bipgraph(ECstar, tmpVlist, bg);
      max_flow(bg, SOURCE, SINK);
      check_flow_in_graph(bg);
      tmpcg = (Graph) malloc(sizeof(struct _Graph)); 
      tmpcg->nodes_list = NULL;
      tmpcg->arcs_list = NULL;
      bipg_to_causalg(bg, tmpcg, ECstar);
    } else {
      
      /**********************************
       * Algorithm STEP 2.1 ELSE part a *
       **********************************/
      VERBMSG("Causalito algorithm: step 2.1, ELSE part a.\n");
      /* compute GprimeC0 */
      GprimeC0 = (Graph) malloc(sizeof(struct _Graph));
      GprimeC0->nodes_list = NULL;
      GprimeC0->arcs_list = NULL;

      ngc0 = g_slist_length(GC0->arcs_list);
      for (i = 0 ; i < ngc0 ; i++) {
	unarc = g_slist_nth_data(GC0->arcs_list, i);
	for (j = 0 ; j < model->neqs ; j++) {
	  e = g_slist_nth_data(model->eqs_list, j);
	  if (e->matched &&
	      !strcmp((unarc->start)->name, e->matched) &&
	      !is_string_in_list(e->eq, EC)) {
	    add_new_arc_to_graph((unarc->start)->name,
				 (unarc->end)->name,
				 GprimeC0);
	    j = model->neqs;
	  }
	}
      }
      DEBUG("Affichage de GprimeC0.\n");
      debug_print_graph(GprimeC0);

      /* compute GammaVprimeC */
      GammaVprimeC = NULL;
      GammaVprimeC = search_successors(GammaVprimeC, GprimeC0, VprimeC);
      DEBUG("GammaVprimeC = %s\n", sprint_char_list(tmpstr,
						    GammaVprimeC));
      /* label the arcs entering GammaVprimeC U VC with C */
      GammaVprimeCUVC = NULL;
      GammaVprimeCUVC = list_concat_uniq(GammaVprimeCUVC, GammaVprimeC, VC);
      ngc0 = g_slist_length(GC0->arcs_list);
      for (i = 0 ; i < ngc0 ; i++) {
	a = g_slist_nth_data(GC0->arcs_list, i);
	if (is_string_in_list((a->end)->name, GammaVprimeCUVC)) {
	  if (strlen(a->label) == 0)
	    sprintf(a->label, "(");
	  else sprintf(a->label, "%s (", a->label);
	  sprint_char_list(tmpstr, C);
	  sprintf(a->label, "%s%s)", a->label, tmpstr);
	}
      }
      
      /**********************************
       * Algorithm STEP 2.1 ELSE part b *
       **********************************/
      VERBMSG("Causalito algorithm: step 2.1, ELSE part a.\n");
      /* compute EGamma */
      EGamma = NULL;
      for (i = 0 ; i < model->neqs ; i++) {
	e = g_slist_nth_data(model->eqs_list, i);
	for (j = 0 ; j < e->nvars ; j++) {
	  ptrstr = g_slist_nth_data(e->vars, j);
	  if (is_string_in_list(ptrstr, GammaVprimeCUVC) &&
	      !is_string_in_list(e->eq, EGamma)) {
	    EGamma = g_slist_append(EGamma, e->eq);
	  }
	}
      }
      
      /**********************************
       * Algorithm STEP 2.1 ELSE part c *
       **********************************/
      VERBMSG("Causalito algorithm: step 2.1, ELSE part a.\n");
      /* search perfect matching */
      tmpVlist = NULL;
      tmpVlist = list_concat_uniq(tmpVlist, GammaVprimeCUVC, VN);

      bg = (Graph) malloc(sizeof(struct _Graph));
      bg->nodes_list = NULL;
      bg->arcs_list = NULL;

      i = g_slist_length(EGamma); 
      j = g_slist_length(tmpVlist);
      if (i != j) {
	BUG("It's a bad idea to compute a perfert matching between %d equations and %d variables. Causalito cannot deal with not self-contained systems.\n", i, j);
      }
      make_bipgraph(EGamma, tmpVlist, bg);
      max_flow(bg, SOURCE, SINK);
      check_flow_in_graph(bg);
      tmpcg = (Graph) malloc(sizeof(struct _Graph)); 
      tmpcg->nodes_list = NULL;
      tmpcg->arcs_list = NULL;
      bipg_to_causalg(bg, tmpcg, ECstar);
    }
    
    /**********************
     * Algorithm STEP 2.2 *
     **********************/
    VERBMSG("Causalito algorithm: step 2.2.\n");

    /* Label the arcs of the new causal graph with Cstar and add them
       to GC0 */
    
    ntmpcg = g_slist_length(tmpcg->arcs_list);
    for (i = 0 ; i < ntmpcg ; i++) {
      a1 = g_slist_nth_data(tmpcg->arcs_list, i);
      a = add_new_arc_to_graph(a1->start->name, a1->end->name, GC0);
      if (strlen(a->label) == 0)
	sprintf(a->label, "(");
      else sprintf(a->label, "%s (", a->label);
      sprint_char_list(tmpstr, Cstar);
      sprintf(a->label, "%s%s)", a->label, tmpstr);
    }
  
  }
  
  VERBMSG("Computing causal graph... Done.\n");
  debug_print_graph(GC0);
  return SUCCESS;
}

/**
 * make_bipgraph: Derive a bipartite graph from the list of equations
 * and variables in the configuration.
 *
 * @mbE: List of equations.
 * @mbV: List of variables.
 * @g: The bipartite graph to compute.
 * 
 * 
 * 
 * Return value: SUCCESS
 **/
int make_bipgraph(GSList * mbE, GSList * mbV, Graph mbG)
{
  char s[2048], * v;
  Equation e;
  GSList * l;
  int mbEn, mbVn, mbEi, mbVi;
  Arc mbback, mbse, mbev, mbvs;
  
  VERBMSG("Creating bipartite graph...\n");
  VERBMSG("Left  nodes: %s\n", sprint_char_list(s, mbE));
  VERBMSG("Right nodes: %s\n", sprint_char_list(s, mbV));

  /*
   * ALGO 
   *
   * create 'back', the arc from SINK to SOURCE, capa = 1000, flow = 0
   * for each 'e' in 'mbE'
   *     'se' = new arc from SOURCE to 'e' capa=1, flow=0
   *     for each 'v' in 'mbV'
   *       if 'v' is in 'e'
   *         'ev' = new arc from 'e' to 'v'
   *         'vs' = arc from 'v' to SINK
   *         if 'vs' does not exist, add it capa=1, flow=0
   *         if 'v' is forced in 'e'
   *             set flow of 'se' = 1
   *             set flow of 'ev' = 1
   *             set flow of 'vs' = 1
   *             add 1 to flow of 'back'
   */

  mbback = add_new_arc_to_graph(SINK, SOURCE, mbG);
  set_arc_capacity(mbback, 1000);
  set_arc_flow(mbback, 0);
  mbEn = g_slist_length(mbE);
  mbVn = g_slist_length(mbV);
  for (mbEi = 0 ; mbEi < mbEn ; mbEi++) {
    l = model->eqs_list;
    while (l != NULL) {
      e = l->data;
      if (!strcmp(e->eq, g_slist_nth_data(mbE, mbEi)))
	l = NULL;
      else
	l = l->next;
    }
    mbse = add_new_arc_to_graph(SOURCE, e->eq, mbG);
    set_arc_capacity(mbse, 1);
    set_arc_flow(mbse, 0);
    for (mbVi = 0 ; mbVi < mbVn ; mbVi++) {
      v = g_slist_nth_data(mbV, mbVi);
      if (is_string_in_list(v, e->vars)) {
	mbev = add_new_arc_to_graph(e->eq, v, mbG);
	set_arc_capacity(mbev, 1);
	set_arc_flow(mbev, 0);
	mbvs = is_arc_in_graph(v, SINK, mbG);
	if (!mbvs) {
	  mbvs = add_new_arc_to_graph(v, SINK, mbG);
	  set_arc_capacity(mbvs, 1);
	  set_arc_flow(mbvs, 0);
	}
	if (e->varf && !strcmp(v, e->varf)) {
	  set_arc_flow(mbse, 1);
	  set_arc_flow(mbev, 1);
	  set_arc_flow(mbvs, 1);
	  mbback->flow++;
	}
      }
    }
  }
  
  VERBMSG("Creating bipartite graph... Done.\n");
  return SUCCESS;
}

/**
 * bipg_to_causalg: computes a causal graph from a bipartite graph
 * @bc_bipg: the input bipartite graph
 * @bc_causg: the causal graph to compute
 * @EE: the list of equations in the bipartite graph
 *
 * ALGO: for each 'bc_eq' in 'bc_equations'
 *         search matched var 'bc_matched_var'
 *         for each var 'eq_var' in 'bc_eq' with 'eq_var' != 'bc_matched_var'
 *           add causal arc 'eq_var' --> 'bc_matched_var'
 * 
 * Return value: SUCCEESS
 **/
int bipg_to_causalg(Graph bc_bipg, Graph bc_causg, GSList * bc_equations)
{
  int n_bc_equations, i_bc_equations, n_bipg_arcs, i_bipg_arcs, i;
  Arc unarc;
  GSList * desarcs;
  char * bc_eq, * bc_matched_var, * unevar;
  Equation eee;

  n_bc_equations = g_slist_length(bc_equations);
  desarcs = bc_bipg->arcs_list;
  n_bipg_arcs = g_slist_length(desarcs);
  for (i_bc_equations = 0 ;
       i_bc_equations < n_bc_equations ;
       i_bc_equations++) {
    /* for each var 'bc_eq', search in the bipartite graph an arc
       whose start is 'bc_eq' and flow is not nul ; the end of the arc
       is the matched var */
    bc_eq = g_slist_nth_data(bc_equations, i_bc_equations);
    for (i_bipg_arcs = 0 ; i_bipg_arcs < n_bipg_arcs ; i_bipg_arcs++) {
      unarc = g_slist_nth_data(desarcs, i_bipg_arcs);
      if (!strcmp((unarc->start)->name, bc_eq) && unarc->flow != 0) {
	bc_matched_var = (unarc->end)->name;
	i_bipg_arcs = n_bipg_arcs;
      }
    }
/*     for (Ai = 0 ; Ai < An ; Ai++) { */
/*       a = g_slist_nth_data(arcs, Ai); */
/*       if (!strcmp((a->start->name), e) && strcmp((a->end)->name, vm)) { */
/* 	add_new_arc_to_graph((a->end)->name, vm, cgt); */
/*       } */
/*     } */
    eee = is_eq_in_model(bc_eq);
    for (i = 0 ; i < eee->nvars ; i++) {
      unevar = g_slist_nth_data(eee->vars, i);
      if (strcmp(unevar, bc_matched_var))
	add_new_arc_to_graph(unevar, bc_matched_var, bc_causg);
    }
  }
  return SUCCESS;
}

/**
 * get_conditions_in_nth_mode: returns the list of conditions in this mode
 * @n: the number of the mode
 * @condsm: the list of conditions
 * 
 * 
 * 
 * Return value: 'condsm'
 **/
GSList * get_conditions_in_nth_mode(GSList * condsm, int nm)
{
  int j;

  for (j = (nm-1) * model->nconds ; j < nm*model->nconds ; j++) {
    condsm = g_slist_append(condsm, g_slist_nth_data(model->confs_list, j));
  }
  return condsm;
}

/**
 * get_equations_with_conditions: compute the list of equations
 * submitted to a condition appearing in the list 'condswc'.
 *
 * @condswc: the list of conditions
 * @eqlc: the result
 * 
 * 
 * 
 * Return value: 'eqlc'
 **/
GSList * get_equations_with_conditions(GSList * eqlc, GSList * condswc)
{
  int i;
  Equation e;
  
  for (i = 0 ; i < model->neqs ; i++) {
    e = g_slist_nth_data(model->eqs_list, i);
    if (is_string_in_list(e->cond, condswc))
      eqlc = g_slist_append(eqlc, e->eq);
  }
  return eqlc;
}
/**
 * get_variables_in_equation: get all the variables appearing in the
 * equations.
 *
 * @varelsg: the computed list of variables.
 * @eqlsg: the list of equations.
 * 
 * 
 * 
 * Return value: 'varelsg', list of variables.
 **/
  
GSList * get_variables_in_equations(GSList * varelsg, GSList * eqlsg)
{
  int i, j, n;
  Equation e;
  char * s;
  
  for (i = 0 ; i < model->neqs ; i++) {
    e = g_slist_nth_data(model->eqs_list, i);
    if (is_string_in_list(e->eq, eqlsg)) {
      n = g_slist_length(e->vars);
      for (j = 0 ; j < n ; j++) {
	s = g_slist_nth_data(e->vars, j);
	if (!is_string_in_list(s, varelsg))
	  varelsg = g_slist_append(varelsg, s);
      }
    }
  }
  return varelsg;
}



/**
 * remember_matched_var: Add to each equation the matched variable in
 * the bipartite graph.
 *
 * @E: List of equations.
 * @Vnotused: 
 * @g: The bipartite graph with maximal flow.
 * 
 * Return value: SUCCESS
 **/
int remember_matched_var(GSList * E, GSList * Vnotused, Graph g)
{
  /*
   * For each equation 'e' of 'E', we search the matched variable in
   * the bipartite graph and write it in the corresponding equation of
   * the model. This is useful later to compute all the stuff
   * appearing in the Causalito algorithm. Note: 'V' is useless here,
   * isn't it ?
   */

  int Ei, i, Vi, j;
  char * v;
  Equation e;
  GSList * V;
  Arc a;

  Ei = g_slist_length(model->eqs_list); 
  for (i = 0 ; i < Ei ; i++) {
    e = g_slist_nth_data(model->eqs_list, i);
    if (is_string_in_list(e->eq, E)) {
      V = e->vars;
      Vi = g_slist_length(V);
      if (Vi != e->nvars) {
	BUG("length of variables list != number of variables\n");
      }
      a = NULL;
      for (j = 0 ; j < Vi && a == NULL ; j++) {
	v = g_slist_nth_data(e->vars, j);
	a = is_arc_in_graph(e->eq, v, g);
	if (a == NULL) {
	  BUG("No arc between equation %s and variable %s\n", e->eq, v);
	}
	if (a->flow == 1) { /* 'v' is the matching var in 'e' */
	  e->matched = v;
	} else { /* 'v' is not the matching var in 'e' */
	  a = NULL; /* set 'a' to NULL else the loop will stop */
	}
      } /* end 'for j' loop */
    }
  }
  return SUCCESS;
}

/**
 * check_flow_in_graph: for each node in the graph, we check that the
 * sum of its incoming flows is the same as the sum of the outgoing
 * ones.
 *
 * @g: the graph to check for
 * 
 * 
 * 
 * Return value: 
 **/
int check_flow_in_graph(Graph gt)
{
  int n, a, i, j, incomingflow, outgoingflow;
  GSList * nl, * al;
  Node N;
  Arc A;
  
  nl = gt->nodes_list;
  al = gt->arcs_list;
  n = g_slist_length(nl);
  a = g_slist_length(al);

  for (i = 0 ; i < n ; i++) {
    N = g_slist_nth_data(nl, i);
    incomingflow = 0;
    outgoingflow = 0;
    for (j = 0 ; j < a ; j++) {
      A = g_slist_nth_data(al, j);
      if (!strcmp((A->start)->name, N->name))
	outgoingflow += A->flow;
      if (!strcmp((A->end)->name, N->name))
	incomingflow += A->flow;
    }
    if (outgoingflow != incomingflow) {
      BUG("Node %s: incoming flow (%d) != outgoing flow (%d)\n", N->name,
	  incomingflow, outgoingflow);
    }
  }
  return SUCCESS;
}
	   
	   
	   
    

/*** Local Variables: ***/
/*** change-log-default-name: "../ChangeLog" ***/
/*** End: ***/

