#include <stdio.h>
#include <glib.h>
#include <string.h>
#include "causalito.h"
#include "model.h"
#include "misc.h"
#include "globals.h"

/* int is_string_in_list(char * s, GSList * l) */
/* { */
/*   return */
/*     g_slist_find_custom(l, s, (GCompareFunc) strcmp) ? SUCCESS : FAILURE; */
/* } */

/*
 * Given a list of conditions, sort all the possible configurations
 * following Gray's code (with an capital G like in Frank Gray :-))
 */


/**
 * compute_Gray_code:
 * @l: 
 * 
 * 
 * 
 * Return value: 
 **/
int compute_Gray_code(GSList *l)
{
  unsigned int gray[1024], n, last_idx = 1, i, j, mask, iter, masks[10];
  char tampon[10];
  
  n = g_slist_length(l);
  if (n != 0) {
    model->nconfs = 1 << n;
    gray[0] = 1;
    gray[1] = 0;
    for (iter = 1 ; iter < n ; iter++) {
      mask = ~((~0) << (iter + 1));
      for (i = 0 ; i <= last_idx ; i++) {
	gray[(~i) & mask] = gray[i];
      }
      for (i = 0 ; i <= last_idx ; i++)
	gray[i] = gray[i] + (1 << iter);
      last_idx = last_idx + (1 << iter);
    }
    for (i = 0 ; i < model->nconds ; i++)
      masks[i] = (1 << i);
    for (i = 0 ; i < model->nconfs ; i++) {
      for (j = 0 ; j < model->nconds ; j++) {
	if (masks[j] & gray[i]) {
	  model->confs_list =
	    g_slist_append(model->confs_list,
			   strdup(g_slist_nth_data(model->conds_name_list,
						   j)));
	} else {
	  strcpy(tampon, "-");
	  strcat(tampon, g_slist_nth_data(model->conds_name_list, j));
	  model->confs_list =
	    g_slist_append(model->confs_list, strdup(tampon));
	}	  
      }
    }
  } else { /* No conditions: this means all equations are always
	      'true' so there is only one mode */
    model->nconfs = 1;
  }
  return SUCCESS;
}

/**
 * sprint_char_list:
 * @L: 
 * @s: 
 * 
 * 
 * 
 * Return value: 
 **/
char * sprint_char_list(char * s, GSList * L)
{
  int l, i;
  
  l = g_slist_length(L);
  s[0] = '\0';
  if (l == 1)
    sprintf(s, "%s", (char *) g_slist_nth_data(L, 0));
  for (i = 1 ; i < l ; i++)
    sprintf(s, "%s %s", s, (char *) g_slist_nth_data(L, i));
  return s;
}

/**
 * remove_elts_from: Put in 'result' all the elements from 'orig' that
 * does not appear in 'toremove'
 *
 * @orig: the initial list
 * @toremove: the list of elements who have to be removed from initial list
 * @result: 'orig' is not modified, so the result is copied to 'result'
 * 
 * 
 * 
 * Return value: result
 **/

GSList * remove_elts_from(GSList * orig, GSList * toremove, GSList * result)
{
  int l, i;
  char * elt;
  
  l = g_slist_length(orig);
  for (i = 0 ; i < l ; i++) {
    elt = g_slist_nth_data(orig, i);
    /* if 'elt' is not in 'toremove', add it to 'result' */
    if (!is_string_in_list(elt, toremove)) /* CAUTION: needs '!' or not ? */
      result = g_slist_append(result, elt);
  }
  return result;
}

/**
 * list_concat_uniq: concat the lists 'list1' and 'list2' in the list 'result' and checks that there is no duplicate elements in the 'result' list
 *
 * @result: @list1: @list2:
 * 
 * 
 * 
 * Return value: 
 **/
GSList * list_concat_uniq(GSList * result, GSList * list1, GSList * list2)
{
  int n, i;
  char * elt;
  
  n = g_slist_length(list1);
  for (i = 0 ; i < n ; i++) {
    elt = g_slist_nth_data(list1, i);
    if (!is_string_in_list(elt, result))
      result = g_slist_append(result, elt);
  }
  n = g_slist_length(list2);
  for (i = 0 ; i < n ; i++) {
    elt = g_slist_nth_data(list2, i);
    if (!is_string_in_list(elt, result))
      result = g_slist_append(result, elt);
  }
  return result;
}

/**
 * list_inter_uniq: return the intersection of two lists and checks
 * for duplicates elements
 *
 * @result: @l1: @l2:
 * 
 * 
 * 
 * Return value: 
 **/
GSList * list_inter_uniq(GSList * result, GSList * list1, GSList * list2)
{
  int n, i;
  char * elt;
  
  n = g_slist_length(list1);
  for (i = 0 ; i < n ; i++) {
    elt = g_slist_nth_data(list1, i);
    if (is_string_in_list(elt, list2) && !is_string_in_list(elt, result))
      result = g_slist_append(result, elt);
  }
  return result;
}

/* int print_char_list(GSList * L, FILE * d) */
/* { */
/*   int l, i; */

/*   l = g_slist_length(L); */
/*   for (i = 0 ; i < l ; i++) */
/*     fprintf(d, "%s ", (char *) g_slist_nth_data(L, i)); */
/*   return 0; */
/* } */

/*** Local Variables: ***/
/*** change-log-default-name: "../ChangeLog" ***/
/*** End: ***/
